////////////////////////////////////////////////////////////////////////////////
// EndDevice.c
//
// 2014/05/27 0.0 Takashi Matsuoka
//

#include <jendefs.h>				// JN516x
#include <AppHardwareApi.h>			// JN516x

#include "config.h"

// Select ToCoNet modules.
// -> ToCoNet_SDK_manual_20140404.pdf P86.
#define ToCoNet_USE_MOD_NWK_LAYERTREE
#define ToCoNet_USE_MOD_NBSCAN
#define ToCoNet_USE_MOD_NBSCAN_SLAVE
#define ToCoNet_USE_MOD_CHANNEL_MGR

#include "ToCoNet.h"				// ToCoNet
#include "ToCoNet_mod_prototype.h"	// ToCoNet
#include "ToCoNet_event.h"			// ToCoNet

#include "utils.h"					// ToCoNetUtils
#include "serial.h"					// ToCoNetUtils

#include <string.h>
#include "i2c.h"

static tsFILE uart;
static tsToCoNet_Nwk_Context* NwkContext;

static uint16 crc16(uint8* ptr, int len)
{
	uint16 crc = 0xffff;
	int i;

	while (len--)
	{
		crc ^= *ptr++;
		for (i = 0; i < 8; i++)
		{
			if (crc & 0x01)
			{
				crc >>= 1;
				crc ^= 0xa001;
			}
			else
			{
				crc >>= 1;
			}
		}
	}

	return crc;
}

static void vProcessEvCore(tsEvent *pEv, teEvent eEvent, uint32 u32evarg);

static void InitializeHardware(bool_t fWarmStart)
{
	// Brown outを設定。
	if (!fWarmStart)
	{
		vAHI_BrownOutConfigure(0, FALSE, FALSE, FALSE, FALSE);
	}

	// UARTを設定。
	static uint8 txBuffer[128];
	static uint8 rxBuffer[128];

	tsSerialPortSetup serial;
	serial.pu8SerialRxQueueBuffer = rxBuffer;
	serial.pu8SerialTxQueueBuffer = txBuffer;
	serial.u32BaudRate = 115200;
	serial.u16AHI_UART_RTS_LOW = 0xffff;
	serial.u16AHI_UART_RTS_HIGH = 0xffff;
	serial.u16SerialRxQueueSize = sizeof (rxBuffer);
	serial.u16SerialTxQueueSize = sizeof (txBuffer);
	serial.u8SerialPort = E_AHI_UART_0;
	serial.u8RX_FIFO_LEVEL = E_AHI_UART_FIFO_LEVEL_1;
	SERIAL_vInit(&serial);

	// vfPrintf()用tsFILEを設定。
	uart.bPutChar = SERIAL_bTxChar;
	uart.u8Device = serial.u8SerialPort;

	// i2cを設定。
	i2c_init(I2C_SPEED_100KHZ);
}

////////////////////////////////////////
// 電源投入時コールバック関数
//
void cbAppColdStart(bool_t bAfterAhiInit)
{
	if (!bAfterAhiInit)
	{
		ToCoNet_REG_MOD_ALL();
	}
	else
	{
		InitializeHardware(FALSE);

		// ユーザ定義イベント処理関数を登録。
		ToCoNet_Event_Register_State_Machine(vProcessEvCore);

		// sToCoNet_AppContextを設定。
		// -> ToCoNet_SDK_manual_20140404.pdf P89.
		sToCoNet_AppContext.u32AppId = APP_ID;
		sToCoNet_AppContext.u32ChMask = CH_MASK;
		sToCoNet_AppContext.u8TxMacRetry = 1;
		sToCoNet_AppContext.bRxOnIdle = TRUE;
	}
}

////////////////////////////////////////
// スリープ復帰時コールバック関数
//
void cbAppWarmStart(bool_t bAfterAhiInit)
{
	if (!bAfterAhiInit)
	{
	}
	else
	{
		InitializeHardware(TRUE);
	}
}

////////////////////////////////////////
// ハードウェア割り込みハンドラ
//
uint8 cbToCoNet_u8HwInt(uint32 u32DeviceId, uint32 u32ItemBitmap)
{
	return FALSE;
}

////////////////////////////////////////
// 送信完了コールバック関数
//
void cbToCoNet_vTxEvent(uint8 u8CbId, uint8 bStatus)
{
}

////////////////////////////////////////
// ネットワークイベントコールバック関数
//
void cbToCoNet_vNwkEvent(teEvent eEvent, uint32 u32arg)
{
	switch (eEvent)
	{
	case E_EVENT_TOCONET_NWK_START:
		ToCoNet_Event_Process(eEvent, u32arg, vProcessEvCore);
		break;
	}
}

////////////////////////////////////////
// ユーザ定義イベント処理関数
//
static void vProcessEvCore(tsEvent *pEv, teEvent eEvent, uint32 u32evarg)
{
	static uint8 sequence = 0;

	switch (pEv->eState)
	{
	case E_STATE_IDLE:
		switch (eEvent)
		{
		case E_EVENT_NEW_STATE:
			vfPrintf(&uart, LB"[E_STATE_IDLE]");
			break;
		case E_EVENT_START_UP:
			vfPrintf(&uart, LB"<E_EVENT_START_UP>");
			{
				tsToCoNet_NwkLyTr_Config nwkConfig;
				memset(&nwkConfig, 0, sizeof (nwkConfig));
				nwkConfig.u8Role = TOCONET_NWK_ROLE_ROUTER;
				nwkConfig.u8Layer = 1;
				NwkContext = ToCoNet_NwkLyTr_psConfig(&nwkConfig);
				if (!NwkContext)
				{
					vfPrintf(&uart, LB"ERROR: Network configuration.");
					break;
				}
				if (!ToCoNet_Nwk_bInit(NwkContext))
				{
					vfPrintf(&uart, LB"ERROR: Network initialize.");
					break;
				}
				if (!ToCoNet_Nwk_bStart(NwkContext))
				{
					vfPrintf(&uart, LB"ERROR: Network start.");
					break;
				}
			}
			break;
		case E_EVENT_TOCONET_NWK_START:
			vfPrintf(&uart, LB"<E_EVENT_TOCONET_NWK_START>");
			ToCoNet_Event_SetState(pEv, E_STATE_RUNNING);
			break;
		}
		break;
	case E_STATE_RUNNING:
		switch (eEvent)
		{
		case E_EVENT_NEW_STATE:
			vfPrintf(&uart, LB"[E_STATE_RUNNING]");
			break;
		case E_EVENT_TOCONET_NWK_CHANGE_CHANNEL:
			vfPrintf(&uart, LB"<E_EVENT_TOCONET_NWK_CHANGE_CHANNEL>");
			break;
		case E_EVENT_TICK_SECOND:
			i2c_start(0x5c, I2C_DIRECTION_WRITE);
			i2c_wait_until_finish();
			vWait(800 * 1800 / 1000);	// 800us
			i2c_stop();

			i2c_start(0x5c, I2C_DIRECTION_WRITE);
			i2c_write(0x03);			// Read holding registers
			i2c_write(0x00);			// Starting address
			i2c_write(0x04);			// Quantity of registers
			i2c_stop();

			i2c_wait_until_finish();
			vWait(1500 * 1800 / 1000);	// 1500us

			i2c_start(0x5c, I2C_DIRECTION_READ);
			i2c_wait_until_finish();
			vWait(30 * 1800 / 1000);	// 30us
			uint8 data[8];
			data[0] = i2c_read(I2C_ACK);
			data[1] = i2c_read(I2C_ACK);
			data[2] = i2c_read(I2C_ACK);
			data[3] = i2c_read(I2C_ACK);
			data[4] = i2c_read(I2C_ACK);
			data[5] = i2c_read(I2C_ACK);
			data[6] = i2c_read(I2C_ACK);
			data[7] = i2c_read(I2C_NAK);
			i2c_stop();

			if (data[0] != 0x03) break;
			if (data[1] != 0x04) break;
			if (((uint16)data[7] << 8 | (uint16)data[6]) != crc16(data, 6)) break;

			int16 humidity;
			int16 temperature;
			humidity = (int16)((uint16)data[2] << 8 | (uint16)data[3]);
			temperature = (int16)((uint16)data[4] << 8 | (uint16)data[5]);
			if ((temperature & 0x8000) != 0)
			{
				temperature = -(temperature & 0x7fff);
			}

			vfPrintf(&uart, LB"%u,%d,%d", u32TickCount_ms, temperature, humidity);

			tsTxDataApp tsTx;
			memset(&tsTx, 0, sizeof (tsTx));

			tsTx.u8CbId = 0;
			//tsTx.u8Retry = 0x80;

			tsTx.u8Cmd = TOCONET_PACKET_CMD_APP_DATA;
			//tsTx.bAckReq = FALSE;
			tsTx.u8Seq = sequence++;
			tsTx.u32DstAddr = TOCONET_NWK_ADDR_PARENT;
			tsTx.u32SrcAddr = ToCoNet_u32GetSerial();
			tsTx.u8Len = 4;
			tsTx.auData[0] = U16_LOWER_U8(temperature);
			tsTx.auData[1] = U16_UPPER_U8(temperature);
			tsTx.auData[2] = U16_LOWER_U8(humidity);
			tsTx.auData[3] = U16_UPPER_U8(humidity);

			if (!ToCoNet_Nwk_bTx(NwkContext, &tsTx))
			{
				vfPrintf(&uart, LB"ERROR: Tx.");
				break;
			}
			ToCoNet_Tx_vProcessQueue();

			break;
		}
		break;
	}
}

////////////////////////////////////////
// ハードウェア割り込み遅延実行部コールバック関数
//
void cbToCoNet_vHwEvent(uint32 u32DeviceId, uint32 u32ItemBitmap)
{
}

////////////////////////////////////////
// 受信コールバック関数
//
void cbToCoNet_vRxEvent(tsRxDataApp *pRx)
{
	vfPrintf(&uart, LB);
	vfPrintf(&uart, "%6u,", u32TickCount_ms / 1000);
	vfPrintf(&uart, "%08x,", pRx->u32SrcAddr);
	vfPrintf(&uart, "%3u,", pRx->u8Lqi);
	vfPrintf(&uart, "%3u,", pRx->u8Seq);
	if (pRx->u8Len == 4)
	{
		int16 temperature;
		int16 humidity;
		temperature = (int16)((uint16)pRx->auData[1] << 8 | (uint16)pRx->auData[0]);
		humidity = (int16)((uint16)pRx->auData[3] << 8 | (uint16)pRx->auData[2]);
		vfPrintf(&uart, "%d,%d", temperature, humidity);
	}
}

////////////////////////////////////////
// メインループコールバック関数
//
void cbToCoNet_vMain()
{
}

////////////////////////////////////////////////////////////////////////////////
